#include <stdio.h>
#include <stdlib.h>

#include <getopt.h>		/* for getopt_long */

#include "pixmap.h"
#include "spline.h"
#include "x11.h"

char *cmnd;		/* name this program was invoked by */


void usage(void)
{
    fprintf(stderr, "Usage: %s "
	    "{--truetype|--x11|-f} fontname "
	    "[{--light|-l} color] "
	    "[{--dark|-d} color] "
	    "[{--textcolor|-t} color] "
	    "[{--shadowcolor|-s} color] "
	    "[{--shadowx|-x} offset] "
	    "[{--shadowy|-y} offset] "
	    "[{--border|-b} pixels] "
	    "[{--em|-M} pixels] "
	    "[{--output|-o} filename] "
	    "\n",
	    cmnd);
};

pixmapT *make_shadow(
			pixmapT * bitmap,
			pixelT color,
			int blur
)
{
    pixmapT *pm = alloc_pixmap(bitmap->rows + 2 * blur, bitmap->cols + 2 * blur);
    pixmapT *pm1 = alloc_pixmap(bitmap->rows + 2 * blur, bitmap->cols + 2 * blur);
    int i, j, k, m, n;
    pixelT transcolor;

    if (alpha(color) == 0)
	color = pixel(255, red(color), green(color), blue(color));
    transcolor = pixel(0, red(color), green(color), blue(color));
    for (i = 0; i < bitmap->rows; i++) {
	for (j = 0; j < bitmap->cols; j++) {
	    if (bitmap->pixels[i][j]) {
		pm->pixels[i + blur][j + blur] = color;
	    } else {
		pm->pixels[i + blur][j + blur] = transcolor;
	    }
	}
    }
    for (i = 0; i < blur; i++) {
	pixmapT *aux;

	for (j = 1; j < pm->rows - 1; j++) {
	    for (k = 1; k < pm->cols - 1; k++) {
		int sum = 0;
		for (m = -1; m <= 1; m++) {
		    for (n = -1; n <= 1; n++) {
			sum += alpha(pm->pixels[j + m][k + n]);
		    }
		}
		sum /= 9;
		pm1->pixels[j][k] =
		    pixel(sum,
			  red(pm->pixels[j][k]),
			  green(pm->pixels[j][k]),
			  blue(pm->pixels[j][k]));
	    }
	}
	aux = pm;
	pm = pm1;
	pm1 = aux;
    }
    free(pm1);
    return pm;
}

pixmapT *make_text(
		      pixmapT * bitmap,
		      pixelT color,
		      int style
)
{
    pixmapT *pm = alloc_pixmap(bitmap->rows, bitmap->cols);
    int i, j;

    if (alpha(color) == 0)
	color = pixel(255, red(color), green(color), blue(color));
    for (i = 0; i < bitmap->rows; i++) {
	for (j = 0; j < bitmap->cols; j++) {
	    if (bitmap->pixels[i][j])
		pm->pixels[i][j] = color;
	}
    }
    return pm;
}



int main(int argc, char *argv[])
{

    int i, j;

    char *fn = NULL;
    int pixelsPerEm = 32;
    pixmapT *bitmap = NULL;
    int rows;
    int cols;
    char *text;

    static struct option o[] =
    {
	{"truetype", 1, 0, 'f'},
	{"light", 1, 0, 'l'},
	{"dark", 1, 0, 'd'},
	{"textcolor", 1, 0, 't'},
	{"x11", 1, 0, 'f'},
	{"shadowcolor", 1, 0, 's'},
	{"shadowx", 1, 0, 'x'},
	{"shadowy", 1, 0, 'y'},
	{"border", 1, 0, 'b'},
	{"em", 1, 0, 'M'},
	{"output", 1, 0, 'o'},
	{NULL, 0, 0, 0}
    };
    int c;
    int longind;
    int use_truetype = 0;
    int use_x11 = 0;
    int border = 0;
    char *p;
    pixelT lightcolor = 0xFFFFFF,
           darkcolor = 0x808080,
           textcolor = 0,
           shadowcolor = 0;
    char *outfilename = "-";
    int shadowx = 0, shadowy = 0;
    pixmapT *shadow, *pmText, *button;
    int rl, gl, bl, rd, gd, bd;

    cmnd = argv[0];

    while ((c = getopt_long(argc, argv, "f:l:d:o:t:s:x:y:b:M:", o, &longind)) != EOF) {
	switch (c) {
	case 0:
	    fn = optarg;
	    switch (o[longind].name[0]) {
	    case 't':
		use_truetype = 1;
		break;
	    case 'x':
		use_x11 = 1;
		break;
	    default:
		printf("%s?\n", o[longind].name);
                usage();
	    }
	    break;

	case 'b':
	    border = strtol(optarg, &p, 0);
	    if (p == optarg || *p)
		usage();
	    break;

	case 'd':
	    if (optarg[0] == '#') {
		darkcolor = strtol(optarg + 1, &p, 16);
		if (p == optarg + 1 || *p)
		    usage();
	    } else {
		/* named colors not yet implemented */
		usage();
	    }
	    break;

	case 'l':
	    if (optarg[0] == '#') {
		lightcolor = strtol(optarg + 1, &p, 16);
		if (p == optarg + 1 || *p)
		    usage();
	    } else {
		/* named colors not yet implemented */
		usage();
	    }
	    break;

	case 's':
	    if (optarg[0] == '#') {
		shadowcolor = strtol(optarg + 1, &p, 16);
		if (p == optarg + 1 || *p)
		    usage();
	    } else {
		/* named colors not yet implemented */
		usage();
	    }
	    break;

	case 't':
	    if (optarg[0] == '#') {
		textcolor = strtol(optarg + 1, &p, 16);
		if (p == optarg + 1 || *p)
		    usage();
	    } else {
		/* named colors not yet implemented */
		usage();
	    }
	    break;

	case 'f':
	    fn = optarg;
	    if (strlen(fn) > 4 && strcmp(fn + strlen(fn) - 4, ".ttf") == 0) {
		use_truetype = 1;
	    } else {
		use_x11 = 1;
	    }
	    break;

	case 'M':
	    pixelsPerEm = strtol(optarg, NULL, 0);
	    break;

	case 'o':
	    outfilename = optarg;
	    break;

	case 'x':
	    shadowx = strtol(optarg, &p, 0);
	    if (p == optarg || *p)
		usage();
	    break;

	case 'y':
	    shadowy = strtol(optarg, &p, 0);
	    if (p == optarg || *p)
		usage();
	    break;

	}
    }

    text = argv[optind];

    /* This is the interesting part, all the rest is just junk to be
     able to show things on the screen */

    if (use_truetype) {
	bitmap = ttfRenderString(fn, text, pixelsPerEm);
    } else if (use_x11) {
	bitmap = x11RenderString(fn, text);
    } else {
	usage();
    }

    rows = bitmap->rows + 2 * border + shadowy + (shadowx || shadowy || shadowcolor) * 4;
    cols = bitmap->cols + 2 * border + shadowx + (shadowx || shadowy || shadowcolor) * 4;

    button = alloc_pixmap(rows, cols);
    if (border) {
	for (i = 0; i < border; i++) {
	    for (j = 0; j < cols - i; j++) {
		button->pixels[i][j] = lightcolor;
		button->pixels[rows - 1 - i][cols - 1 - j] = darkcolor;
	    }
	    for (j = 0; j < rows - i; j++) {
		button->pixels[j][i] = lightcolor;
		button->pixels[rows - 1 - j][cols - 1 - i] = darkcolor;
	    }
	}
    }
    rl = red(lightcolor);
    gl = green(lightcolor);
    bl = blue(lightcolor);
    rd = red(darkcolor);
    gd = green(darkcolor);
    bd = blue(darkcolor);
    for (i = border; i < rows - border; i++) {
	for (j = border; j < cols - border; j++) {
	    int r, g, b;
	    double scale = (double) (i + j) / (rows + cols);
	    r = rl + scale * (rd - rl);
	    g = gl + scale * (gd - gl);
	    b = bl + scale * (bd - bl);
	    button->pixels[i][j] = pixel(255, r, g, b);
	}
    }
    if (shadowx || shadowy || shadowcolor) {
	shadow = make_shadow(bitmap, shadowcolor, 2);
	pixmap_replace(button, shadow, border + shadowx, border + shadowy);
    }
    pmText = make_text(bitmap, textcolor, 0);
    pixmap_replace(button, pmText, border, border);
    write_pixmap(outfilename, button);

    return 0;
}
