#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "ttf.h"
#include "pixmap.h"

int fraction;

pixmapT *ttfRenderString(
    const char *fn /* font file name */ ,
    const char *string	/* the string to render. SGML style numeric
                         * escapes are planned, but not yet implemented
                         */ ,
    int pixelsPerEm	/* Actually, 1 em doesn't really seem to be the
                         * width of am 'M' with most truetype fonts,
                         * but considerably more. Should get a better
                         * metric for scaling (like x-height).
                         */
)
{
    pixmapT *bitmap;
    int last;

    XPoint Points[16][1024];
    int nPoints[16];

    int i, j;

    ttfHandle *handle;
    int charCode;
    int whichGlyph;
    ttfGlyph *glyph;
    ttfRenderList renderList[1024];
    int scale;
    int rows;
    int cols;
    int xoff;
    const char *p;

    /* Open a TrueType font file and parse the most important tables */
    if ((handle = ttfOpen(fn)) == NULL)
	return NULL;

    /* compute scaling */
    fraction = handle->unitsPerEm / pixelsPerEm;
    scale = fraction / 16;

    /* determine size of bitmap */
    cols = 0;
    for (p = string; *p; p++) {
	if (*p == '&' && *(p + 1) == '#') {
	    /* SGML escape */
	    charCode = strtol(p + 2, &p, 10);
	    if (*p != ';') p--;
	} else {
	    charCode = (unsigned char) *p;
	}
	whichGlyph = ttfCode2Glyph(handle, charCode, 3, 1);	/* 3 1 means Unicode */
	cols += handle->hmtxTable[whichGlyph].advanceWidth / fraction;
    }
    rows = (handle->yMax - handle->yMin) / fraction + 1;
    bitmap = alloc_pixmap(rows, cols);

    xoff = 0;
    for (p = string; *p; p++) {
	if (*p == '&' && *(p + 1) == '#') {
	    /* SGML escape */
	    charCode = strtol(p + 2, &p, 10);
	    if (*p != ';') p--;
	} else {
	    charCode = (unsigned char) *p;
	}
	whichGlyph = ttfCode2Glyph(handle, charCode, 3, 1);	/* 3 1 means Unicode */
	/* Read the specified glyph from the font */
	if ((glyph = ttfReadGlyph(handle, whichGlyph)) == NULL) {
	    fprintf(stderr, "readGlyph failed\n");
	    exit(1);
	}
	/* Convert each glyph contour to a polygon */
	last = 0;
	for (i = 0; i < glyph->nContours; i++) {
	    nPoints[i] = ttfLinearize(Points[i],
				      glyph->endPoints[i] - last + 1,
				      glyph->points + last)
		- Points[i];
	    last = glyph->endPoints[i] + 1;
	}

	/* Move everything about a bit to fit in a grid with origo (0,0) */
	for (i = 0; i < glyph->nContours; i++)
	    for (j = 0; j < nPoints[i]; j++) {
		Points[i][j].x = Points[i][j].x;	/* XXX ??? */
		Points[i][j].y = handle->yMax - Points[i][j].y;
	    }

	/* Scan-convert the polygon to a render list (actually a scan-line list */
	memset(renderList, 0, sizeof(renderList));
	for (i = 0; i < glyph->nContours; i++)
	    ttfRender(renderList, nPoints[i], Points[i]);

	/* Rasterize the render-list into a list of pixels to turn on */
	ttfRasterize(bitmap, xoff, renderList);
	xoff += handle->hmtxTable[whichGlyph].advanceWidth / fraction;
    }
    return bitmap;
}


/*
 * return a map of valid chars in the font.
 * Each character is represented by 1 char. 
 * Bit 0 is set if the character is present
 * The last character has bit 1 set.
 */
char *ttfCharMap(const char *fn) {
    ttfHandle *handle;
    char *charmap = malloc(65536);
    int i;
    int lastchar = 0;

    /* Open a TrueType font file and parse the most important tables */
    if ((handle = ttfOpen(fn)) == NULL)
	return NULL;
    
    for (i = 0; i < 65536; i++) {
	int glyph = ttfCode2Glyph(handle, i, 3, 1);
	if (glyph == -1) {
	    charmap[i] = 0;
	} else {
	    charmap[i] = 1;
	    lastchar = i;
	}
    }
    charmap[lastchar] |= 2;

    return charmap;
    
}
