#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "ttf.h"

/************************************************************************/

#define DEBUG_MISC 1
#define DEBUG_OFFSETTABLE 2
#define DEBUG_LOCATION 4
#define DEBUG_HMETRICS 8
#define DEBUG_FORMAT4 0x10

#ifndef DEBUG
#define DEBUG 0
#endif

/***********************************************************************
  Read most of the interesting tables from the TrueType font file.
  Currently, the offset table and the tables relevant to the drawing
  of glyphs are read.

  This would be a good place to insert code for reading the Postscript
  metrics table.

  */

static BOOL readTables(ttfHandle * handle)
{
    int i;
    const char *p;
    int lastAW = 0;

    handle->offsetTable = (ttfOffsetTable *) handle->fileData;

#if DEBUG & DEBUG_OFFSETTABLE
    fprintf(stderr, "Offset Table:\n");
    fprintf(stderr, "  sfntVersion: %lx\n",
	    TTF_Fixed(&handle->offsetTable->sfntVersion));
    fprintf(stderr, "  numTables: %u\n",
	    TTF_CARD16(&handle->offsetTable->numTables));
    fprintf(stderr, "\n");
#endif

    handle->numTables = TTF_CARD16(&handle->offsetTable->numTables);

    handle->tableDirectory = (ttfTableDirectory *)
	(handle->fileData + sizeof(ttfOffsetTable));

    for (i = 0; i < handle->numTables; i++) {
#if DEBUG & DEBUG_OFFSETTABLE
	fprintf(stderr, "  %4.4s", &handle->tableDirectory[i].tag);
	fprintf(stderr, "  offset: %06lx",
		TTF_CARD32(&handle->tableDirectory[i].offset));
	fprintf(stderr, "  length: %06lx",
		TTF_CARD32(&handle->tableDirectory[i].length));
	fprintf(stderr, "\n");
#endif

	switch (TTF_CARD32(&handle->tableDirectory[i].tag)) {
	case TTFTAG('g', 'l', 'y', 'f'):
	    handle->baseGlyphData =
		(handle->fileData + TTF_CARD32(&handle->tableDirectory[i].offset));
	    break;

	case TTFTAG('h', 'e', 'a', 'd'):
	    handle->fontHeader = (ttfFontHeader *)
		(handle->fileData + TTF_CARD32(&handle->tableDirectory[i].offset));
	    handle->unitsPerEm = TTF_CARD16(&handle->fontHeader->unitsPerEm);
	    handle->xMin = TTF_INT16(&handle->fontHeader->xMin);
	    handle->xMax = TTF_INT16(&handle->fontHeader->xMax);
	    handle->yMin = TTF_INT16(&handle->fontHeader->yMin);
	    handle->yMax = TTF_INT16(&handle->fontHeader->yMax);
	    fprintf(stderr, "unitsPerEm = %d\n", handle->unitsPerEm);
	    fprintf(stderr, "bbox = (%d, %d) ... (%d, %d)\n",
		   handle->xMin, handle->yMin,
		   handle->xMax, handle->yMax
		);
	    break;

	case TTFTAG('h', 'h', 'e', 'a'):
	    handle->hhea =
		(handle->fileData + TTF_CARD32(&handle->tableDirectory[i].offset));
	    break;

	case TTFTAG('h', 'm', 't', 'x'):
	    handle->hmtx =
		(handle->fileData + TTF_CARD32(&handle->tableDirectory[i].offset));
	    break;

	case TTFTAG('l', 'o', 'c', 'a'):
	    handle->location =
		(handle->fileData + TTF_CARD32(&handle->tableDirectory[i].offset));
	    break;

	case TTFTAG('m', 'a', 'x', 'p'):
	    handle->maxProfile = (ttfMaxProfile *)
		(handle->fileData + TTF_CARD32(&handle->tableDirectory[i].offset));
	    break;

	case TTFTAG('c', 'm', 'a', 'p'):
	    fprintf(stderr, "found cmap\n");
	    handle->baseCharToIndexMap = (char *)
		(handle->fileData + TTF_CARD32(&handle->tableDirectory[i].offset));
	    fprintf(stderr, "Version: %lx\n", TTF_CARD16(handle->baseCharToIndexMap + 0));
	    handle->nCharMaps = TTF_CARD16(handle->baseCharToIndexMap + 2);
	    fprintf(stderr, "tableCount: %x\n", handle->nCharMaps);
	    break;
	}

    }
#if DEBUG & DEBUG_OFFSETTABLE
    fprintf(stderr, "\n");
#endif

    if (handle->baseGlyphData == NULL) {
	fprintf(stderr, "'glyf' table missing\n");
	return 0;
    }
    if (handle->fontHeader == NULL) {
	fprintf(stderr, "'head' table missing\n");
	return 0;
    }
    if (handle->location == NULL) {
	fprintf(stderr, "'loca' table missing\n");
	return 0;
    }
    if (handle->maxProfile == NULL) {
	fprintf(stderr, "'maxp' table missing\n");
	return 0;
    }
    if (handle->hhea == NULL) {
	fprintf(stderr, "'hhea' table missing\n");
	return 0;
    }
    if (handle->hmtx == NULL) {
	fprintf(stderr, "'hmtx' table missing\n");
	return 0;
    }
    handle->nGlyphs = TTF_CARD16(&handle->maxProfile->numGlyphs);

#if DEBUG & DEBUG_MISC
    fprintf(stderr, "No of glyphs: %d\n\n", handle->nGlyphs);
#endif

    if ((handle->glyphData =
	 malloc(handle->nGlyphs * sizeof(ttfGlyphData *))) == NULL) {
	perror("malloc Glyph Data pointers");
	return 0;
    }
    memset(handle->glyphData, 0, handle->nGlyphs * sizeof(ttfGlyphData *));

#if DEBUG & DEBUG_LOCATION
    fprintf(stderr, "Location Table:\n");
#endif

    switch (TTF_INT16(&handle->fontHeader->indexToLocFormat)) {
    case 0:
	{
	    CARD16 *p = (CARD16 *) handle->location;

#if DEBUG & DEBUG_LOCATION
	    fprintf(stderr, "  Short indexToLocFormat\n");
#endif

	    for (i = 0; i < handle->nGlyphs; i++)
		handle->glyphData[i] = (ttfGlyphData *)
		    (handle->baseGlyphData + 2 * TTF_CARD16(&p[i]));
	    break;
	}

    case 1:
	{
	    CARD32 *p = (CARD32 *) handle->location;

#if DEBUG & DEBUG_LOCATION
	    fprintf(stderr, "  Long indexToLocFormat\n");
#endif

	    for (i = 0; i < handle->nGlyphs; i++)
		handle->glyphData[i] = (ttfGlyphData *)
		    (handle->baseGlyphData + TTF_CARD32(&p[i]));
	    break;
	}

    default:
	fprintf(stderr, "Unknown indexToLocFormat %d\n",
		TTF_INT16(&handle->fontHeader->indexToLocFormat));
	return 1;
    }

#if DEBUG & DEBUG_LOCATION
    fprintf(stderr, "\n");

    for (i = 0; i < handle->nGlyphs; i++)
	fprintf(stderr, "  %d: %d\n", i,
	 (BYTE *) handle->glyphData[i] - (BYTE *) handle->baseGlyphData);

    fprintf(stderr, "\n");
#endif

    if ((handle->charToIndexMap = malloc(sizeof(ttfCharToIndexMap *) * handle->nCharMaps)) == NULL) {
	perror("malloc charToIndexMap pointers");
	return 0;
    }
    for (i = 0; i < handle->nCharMaps; i++) {
	const char *p = handle->baseCharToIndexMap + 4 + 8 * i;
	ttfCharToIndexMap *ctim;
	if ((handle->charToIndexMap[i] = ctim = malloc(sizeof(ttfCharToIndexMap))) == NULL) {
	    perror("malloc charToIndexMap");
	    return 0;
	}
	ctim->platformID = TTF_CARD16(p);
	fprintf(stderr, "platformID: %x\n", ctim->platformID);
	ctim->specificID = TTF_CARD16(p + 2);
	fprintf(stderr, "specificID: %x\n", ctim->specificID);
	ctim->offset = TTF_CARD32(p + 4);
	fprintf(stderr, "offset: %lx\n", ctim->offset);

	p = handle->baseCharToIndexMap + ctim->offset;
	ctim->format = TTF_CARD16(p);
	p += 2;
	fprintf(stderr, "format: %x\n", ctim->format);
	ctim->length = TTF_CARD16(p);
	p += 2;
	fprintf(stderr, "length: %x\n", ctim->length);
	ctim->version = TTF_CARD16(p);
	p += 2;
	fprintf(stderr, "version: %x\n", ctim->version);

	switch (ctim->format) {
	    int j;

	case 4:
	    /* Format 4.
         * This consists of a number of segments, each of which
         * describes a range of characters. Either the glyph number
         * can be obtained for the whole range by adding a constant
         * to the character code (mod 65536), or each character must
         * be looked up in a table.
         *
         * Actually quite simple, unless you deduce the whole shebang
         * from povray sources.
         */
	    ctim->segCount = TTF_CARD16(p) >> 1;
	    p += 2;		/* this count is obviously in bytes, not entries */
	    #if DEBUG & DEBUG_FORMAT4
	    fprintf(stderr, "segCount: %x\n", ctim->segCount);
	    #endif
	    ctim->searchRange = TTF_CARD16(p);
	    p += 2;
	    #if DEBUG & DEBUG_FORMAT4
	    fprintf(stderr, "searchRange: %x\n", ctim->searchRange);
	    #endif
	    ctim->entrySelector = TTF_CARD16(p);
	    p += 2;
	    #if DEBUG & DEBUG_FORMAT4
	    fprintf(stderr, "entrySelector: %x\n", ctim->entrySelector);
	    #endif
	    ctim->rangeShift = TTF_CARD16(p);
	    p += 2;
	    #if DEBUG & DEBUG_FORMAT4
	    fprintf(stderr, "rangeShift: %x\n", ctim->rangeShift);
	    #endif
	    if ((ctim->format4Segs = malloc(sizeof(ttfFormat4Seg) * ctim->segCount)) == NULL) {
		perror("malloc format4Segs");
		return 0;
	    }
	    for (j = 0; j < ctim->segCount; j++) {
		ctim->format4Segs[j].endCount = TTF_CARD16(p);
		p += 2;
	    }
	    p += 2;		/* skip reserved bytes */
	    for (j = 0; j < ctim->segCount; j++) {
		ctim->format4Segs[j].startCount = TTF_CARD16(p);
		p += 2;
	    }
	    for (j = 0; j < ctim->segCount; j++) {
		ctim->format4Segs[j].idDelta = TTF_CARD16(p);
		p += 2;
	    }

	    for (j = 0; j < ctim->segCount; j++) {
		ctim->format4Segs[j].idRangeOffset = TTF_CARD16(p);
		if (ctim->format4Segs[j].idRangeOffset != 0) {
		    const char *p1 = p + ctim->format4Segs[j].idRangeOffset;
		    int k;
		    int n = ctim->format4Segs[j].endCount - ctim->format4Segs[j].startCount + 1;

		    if ((ctim->format4Segs[j].glyphTable = malloc(sizeof(CARD16) * n)) == NULL) {
			perror("malloc format4 glyph table");
			return 0;
		    }
		    for (k = 0;
			 k < n;
			 k++) {
			ctim->format4Segs[j].glyphTable[k] = TTF_CARD16(p1);
			p1 += 2;
		    }
		}
		p += 2;
	    }
#if DEBUG & DEBUG_FORMAT4
	    for (j = 0; j < ctim->segCount; j++) {
		fprintf(stderr, "%d: [%d ... %d] =>",
		       j,
		       ctim->format4Segs[j].startCount,
		       ctim->format4Segs[j].endCount);
		if (ctim->format4Segs[j].idRangeOffset) {
		    int k;
		    int n = ctim->format4Segs[j].endCount - ctim->format4Segs[j].startCount + 1;

		    for (k = 0;
			 k < n;
			 k++) {
			fprintf(stderr, " %d", ctim->format4Segs[j].glyphTable[k]);
		    }
		    fprintf(stderr, "\n");
		} else {
		    fprintf(stderr, " [%d ... %d]\n",
			   (CARD16) (ctim->format4Segs[j].startCount +
				     ctim->format4Segs[j].idDelta),
			   (CARD16) (ctim->format4Segs[j].endCount +
				     ctim->format4Segs[j].idDelta)
			);

		}
	    }
#endif
	    break;
	default:
	    fprintf(stderr, "ignoring unknown format %d\n",
		   handle->charToIndexMap[i]->format);
	}
    }

    /* hhea - horizontal metrics header */
    p = handle->hhea;
    p += 4;			/* skip version */
    handle->ascender = TTF_INT16(p);
    p += 2;
    handle->descender = TTF_INT16(p);
    p += 2;
    handle->linegap = TTF_INT16(p);
    p += 2;
    handle->advanceWidthMax = TTF_CARD16(p);
    p += 2;
    handle->minLeftSideBearing = TTF_INT16(p);
    p += 2;
    handle->minRightSideBearing = TTF_INT16(p);
    p += 2;
    handle->xMaxExtent = TTF_INT16(p);
    p += 2;
    p += 7 * 2;
    handle->metricDataFormat = TTF_INT16(p);
    p += 2;
    handle->numberOfHMetrics = TTF_CARD16(p);
    p += 2;

    /* horizontal metrics */
    if ((handle->hmtxTable = malloc(handle->nGlyphs * sizeof(*handle->hmtxTable))) == NULL) {
	perror("malloc hmtxTable");
	return 0;
    }
    p = handle->hmtx;
    for (i = 0; i < handle->numberOfHMetrics; i++) {
	lastAW = handle->hmtxTable[i].advanceWidth = TTF_CARD16(p);
	p += 2;
	handle->hmtxTable[i].lsb = TTF_INT16(p);
	p += 2;
#if DEBUG & DEBUG_HMETRICS
	fprintf(stderr, "%d: AW = %d, lsb = %d\n", i, lastAW, handle->hmtxTable[i].lsb);
#endif
    }
    for (; i < handle->nGlyphs; i++) {
	lastAW = handle->hmtxTable[i].advanceWidth = lastAW;
	handle->hmtxTable[i].lsb = TTF_INT16(p);
	p += 2;
#if DEBUG & DEBUG_HMETRICS
	fprintf(stderr, "%d: lsb = %d\n", i, handle->hmtxTable[i].lsb);
#endif
    }
    return 1;
}

/***********************************************************************
  Open a TrueType font file and return a handle to it.  This function
  will also read in most intreresting tables from the file.

  (A good, matching function would be a ttfClose, but I've been too
  lazy so far.)

*/

ttfHandle *ttfOpen(const char *fn)
{
    ttfHandle *handle;
    FILE *fp;
    BYTE *fileData;

    if ((handle = malloc(sizeof(ttfHandle))) == NULL) {
	perror("malloc handle");
	return NULL;
    }
    memset(handle, 0, sizeof(ttfHandle));

    if ((fp = fopen(fn, "rb")) == NULL) {
	perror("fopen");
	free(handle);
	return NULL;
    }
    fseek(fp, 0l, SEEK_END);
    handle->fileLength = ftell(fp);
    rewind(fp);

    if ((fileData = malloc(handle->fileLength)) == NULL) {
	perror("malloc data");
	fclose(fp);
	free(handle);
	return NULL;
    }
    /* Poor man's mmap, read everything into memory */
    if (fread(fileData, handle->fileLength, 1, fp) != 1) {
	perror("fread");
	fclose(fp);
	free(handle);
	return NULL;
    }
    handle->fileData = fileData;

    fclose(fp);

    if (!readTables(handle)) {
	fprintf(stderr, "readTables failed\n");
	free(handle);
	return NULL;
    }
    return handle;
}
