#include <assert.h>
#include <X11/Xlib.h>
#include <stdio.h>
#include <stdlib.h>

#include "ttf.h"
#include "pixmap.h"

/***********************************************************************
  Debugging
*/

#define DEBUG_MISC 1

#ifndef DEBUG
#define DEBUG 0
#endif

/***********************************************************************
  Useful macros
*/

/* Remove any old declaration */
#ifdef ABS
#undef ABS
#endif

#define ABS(x) ((x) >= 0 ? (x) : -(x))

/***********************************************************************

   Determine when to stop subdividing

   This one is quite simpleminded, it just checks to see how far apart
   the on-curve points are.

   A better one might be to check if the off-curve point falls on a
   straight line between the on-curve points.

*/

#define TERM(xA, yA, xB, yB, xC, yC) \
	(ABS(xA-xC) <= termval && ABS(yA-yC) <= termval)

/************************************************************************

  ttfBezier - Subdivide a Cubic Bezier into line segments

  XPoint *pt		Where to store the generated points
  return XPoint * 	Pointer to element after last point stored
  int depth		The deepest to go  (to make sure we finish)
  long xA, xB, yC, yC	on-curve points
  long xB, yB		off-curve control point

*/

XPoint *ttfBezier(XPoint * pt,
		  int depth,
		  long xA, long yA,
		  long xB, long yB,
		  long xC, long yC)
{
    long xAB, yAB;
    long xBC, yBC;
    long xABC, yABC;
    long termval = fraction / 2;

    /* Time to quit? */
    if (!depth || TERM(xA, yA, xB, yB, xC, yC)) {
	/* Since the order of subdivision has been carefully chosen, the
       points will be added to the list in the correct order */
	pt->x = xC;
	pt->y = yC;
	pt++;
	return pt;
    }
    /* If all coordinate values can be guaranteed to be positive, all
     divisions can be replaced by shifts */

    xAB = (xA + xB) / 2;	/* AB = midpoint(A, B) */
    yAB = (yA + yB) / 2;

    xBC = (xB + xC) / 2;	/* BC = midpoint(B, C) */
    yBC = (yB + yC) / 2;

    xABC = (xAB + xBC) / 2;	/* ABC = midpoint(AB, BC) */
    yABC = (yAB + yBC) / 2;

    pt = ttfBezier(pt,
		   depth - 1,
		   xA, yA,
		   xAB, yAB,
		   xABC, yABC);

    pt = ttfBezier(pt,
		   depth - 1,
		   xABC, yABC,
		   xBC, yBC,
		   xC, yC);

    return pt;
}

/***********************************************************************

  ttfLinearize - convert a TrueType glyph to a polygon

  XPoint *lines		Where to store the coordinates of the line segments
  return XPoint *lines	Pointer to element after last written
  int nPoints		No of points defining the glyph
  ttfPoint *points	Array of points

*/

XPoint *ttfLinearize(XPoint * lines, int nPoints, ttfPoint * points)
{
    BOOL lastWasOnPoint;
    XPoint onPoint, offPoint;
    XPoint *pl;			/* Output line-segments */
    ttfPoint *pp;		/* Input points defining glyph */

#if DEBUG & DEBUG_MISC
    fprintf(stderr, "Linearizing glyph\n");
#endif

    pl = lines;
    pp = points;

    lastWasOnPoint = 0;

    if (!points[0].flags & TTFPOINT_ON)
	fprintf(stderr, "WARNING! First point of curve was NOT an on point\n");

    /* Initialize with first point which must be an on point */
    pl->x = pp->x;
    pl->y = pp->y;
    pl++;

    onPoint.x = pp->x;
    onPoint.y = pp->y;
    lastWasOnPoint = 1;
    pp++;
    nPoints--;

    while (nPoints--) {
	if (pp->flags & TTFPOINT_ON) {
	    /* This is an on-curve point */

	    if (lastWasOnPoint) {
		/* Two on-curve points in a row constitutes a straight line */
		pl->x = pp->x;
		pl->y = pp->y;
		pl++;
	    } else {
		/* We have a complete bezier, draw it */
		pl = ttfBezier(pl, DEPTH,
			       onPoint.x, onPoint.y,
			       offPoint.x, offPoint.y,
			       pp->x,
			       pp->y);
	    }

	    /* Save the on point for later */
	    onPoint.x = pp->x;
	    onPoint.y = pp->y;
	    lastWasOnPoint = 1;
	} else {
	    /* This is an off-curve point */

	    if (!lastWasOnPoint) {
		/* If the last point wasn't an on-curve point, we have to
	 insert one halfway between the last and current off-curve
	 points */

		int x, y;

		/* Create a new on-curve point */
		x = (offPoint.x + pp->x) / 2;
		y = (offPoint.y + pp->y) / 2;

		/* Draw the bezier */
		pl = ttfBezier(pl, DEPTH,
			       onPoint.x, onPoint.y,
			       offPoint.x, offPoint.y,
			       x,
			       y);

		/* Save this on-curve point for later */
		onPoint.x = x;
		onPoint.y = y;
	    }
	    /* Save the off curve point for later */
	    offPoint.x = pp->x;
	    offPoint.y = pp->y;
	    lastWasOnPoint = 0;
	}

	/* Next control point please */
	pp++;
    }

    /* Close the curve */
    if (!lastWasOnPoint) {
	pl = ttfBezier(pl, DEPTH,
		       onPoint.x, onPoint.y,
		       offPoint.x, offPoint.y,
		       points->x,
		       points->y);
    }
    pl->x = points->x;
    pl->y = points->y;
    pl++;

#if DEBUG & DEBUG_MISC
    fprintf(stderr, "  No of line segments: %d\n", pl - lines);
#endif

    return pl;
}

/***********************************************************************
  This is really ugly code.  I got quite tired, so I didn't bother
  with dynamic allocation of memory or anything fancy.  It should be
  possible to replace this with X's rasterizer.

  What it does is to take a polygon and scan-convert it by walking
  through the list of line segments, and for each line segment
  calculate which scanlines it crosses and where.

  To avoid all special cases where a lines endpoints fall exactly on a
  scan line, all coordinates are offset by 1, unfortunately, this
  isn't allowed by the TrueType specification and will probably make
  some fonts which rely on this look bad.

  The output is a Render List where each line has a list of crossings
  and wether the winding number should increase or decrease.

  Take a look at Graphics Gems I, `Fast Scan Conversion of Arbitrary
  Polygons' for more theory on how it works.

*/

static void addToList(ttfRenderList * renderList, int x, int y, int on)
{
#if 1
    y /= fraction;
#endif
    renderList[y].columns[renderList[y].nColumns] = x;
    renderList[y].on[renderList[y].nColumns] = on;
    renderList[y].nColumns++;
}

void ttfRender(ttfRenderList * renderList, int nLines, XPoint * lines)
{
    int i;

    for (i = 0; i < nLines; i++)
	lines[i].y |= 1;

    for (i = 1; i < nLines; i++) {
	int x, y;
	int dx, dy;

	dx = lines[i].x - lines[i - 1].x;
	dy = lines[i].y - lines[i - 1].y;

	if (dy > 0) {
	    /* Moving down, so the interior is to the left */
	    for (y = ((lines[i - 1].y + fraction / 2) / fraction * fraction) + fraction / 2;
		 y <= lines[i].y;
		 y += fraction) {
		x = lines[i - 1].x + dx * (y - lines[i - 1].y) / dy;
		assert(y / fraction >= 0 && y / fraction < 1024);
		addToList(renderList, x, y, False);	/* add an off point */
	    }
	} else if (dy < 0) {
	    /* Moving up, so the interior is to the right */
	    for (y = ((lines[i - 1].y + fraction / 2) / fraction * fraction) - fraction / 2;
		 y >= lines[i].y;
		 y -= fraction) {
		x = lines[i - 1].x + dx * (y - lines[i - 1].y) / dy;
		assert(y / fraction >= 0 && y / fraction < 1024);
		addToList(renderList, x, y, True);	/* add an on point */
	    }
	} else {
	    /* Horizontal lines are a special case */
	    /* just ignore for now */
	}
    }
}

/***********************************************************************
  Generate a bitmap from the render list.

  This is a quite straightforward function.  Sort the coordinates of
  each line, and start drawing from the left edge.  Start turning on
  pixels after an on-crossing and turning off after an off-crossing.

*/

static int compare(ttfRenderList * rl, int a, int b)
{
    if (rl->columns[a] < rl->columns[b])
	return 1;
    if (rl->columns[a] > rl->columns[b])
	return 0;
    if (rl->on[a] > rl->on[b])
	return 1;
    return 0;
}

void ttfRasterize(
		     pixmapT * bitmap,
		     int xoff,
		     ttfRenderList * renderList)
{
    int i, j, k;
    int rows = bitmap->rows;
    int cols = bitmap->cols;

    for (i = 0; i < rows; i++) {
	int on;
	int x;

	for (j = 1; j < renderList[i].nColumns; j++)
	    for (k = 0; k < j; k++)
		if (compare(&renderList[i], j, k)) {
		    int col = renderList[i].columns[j];
		    int o = renderList[i].on[j];
		    renderList[i].columns[j] = renderList[i].columns[k];
		    renderList[i].on[j] = renderList[i].on[k];
		    renderList[i].columns[k] = col;
		    renderList[i].on[k] = o;
		}
	if (renderList[i].nColumns) {
	    on = 0;
	    x = 0;
	    for (j = 0; j < renderList[i].nColumns; j++) {
		while ((x * fraction + fraction / 2) < renderList[i].columns[j]) {
		    if (on) {
			if (x + xoff < cols) {
			    bitmap->pixels[i][x + xoff] = pixel(255, 0, 0, 0);
			}
		    }
		    x++;
		}

		if (renderList[i].on[j])
		    on++;
		else
		    on--;

		if (on < 0)
		    fprintf(stderr, "BUG! on is negative\n");
	    }
	}
    }

}
